@file:Suppress("UsePropertyAccessSyntax")

package ru.dbotthepony.kommons.util

import java.util.function.Consumer
import java.util.function.Supplier
import kotlin.reflect.KMutableProperty0
import kotlin.reflect.KProperty
import kotlin.reflect.KProperty0

/**
 * [Supplier] and [Consumer] combined in single interface
 */
interface Delegate<V> : Supplier<V>, Consumer<V> {
	class UnboundBox<V> : Delegate<V> {
		private var value: Any? = Companion

		override fun get(): V {
			val get = value
			if (get === Companion) throw UninitializedPropertyAccessException()
			return get as V
		}

		override fun accept(t: V) {
			value = t
		}

		private companion object
	}

	class Box<V>(private var value: V) : Delegate<V> {
		override fun get(): V {
			return value
		}

		override fun accept(t: V) {
			value = t
		}
	}

	class SmartBox<V>(value: V, private val getter: DelegateGetter<V>, private val setter: DelegateSetter<V>) : Delegate<V> {
		private val value = Box(value)

		override fun get(): V {
			return getter(value)
		}

		override fun accept(t: V) {
			setter(value, t)
		}
	}

	class Of<V>(getter: Supplier<V>, setter: Consumer<V>) : Delegate<V>, Supplier<V> by getter, Consumer<V> by setter {
		constructor(prop: KMutableProperty0<V>) : this(prop::get, prop::set)
		@Deprecated("Warning: Using read-only property as delegate")
		constructor(prop: KProperty0<V>) : this(prop::get, Consumer { })
	}
}

operator fun <V> Supplier<V>.getValue(ref: Any?, property: KProperty<*>): V {
	return get()
}

operator fun <V> Consumer<V>.setValue(ref: Any?, property: KProperty<*>, value: V) {
	accept(value)
}

var <V> Delegate<V>.value: V
	get() = get()
	set(value) { accept(value) }
